//  Created by Marco Giobbi on 13/01/14.
/* Copyright (c) 1997-1999 Miller Puckette.
 * For information on usage and redistribution, and for a DISCLAIMER OF ALL
 * WARRANTIES, see the file, "LICENSE.txt," in this distribution.  */

/*  "filters", both linear and nonlinear.
 */

#ifdef linux
#define _GNU_SOURCE
#endif

#include "m_pd.h"
#include <math.h>
#include <stdio.h>
#define PI 3.14159265358979323846264338328

//Derived from biquad~ function in d_filter.c
/* ---------------- biquad~ - raw biquad filter ----------------- */

typedef struct biquadctl
{
    t_sample c_x1;
    t_sample c_x2;
    t_sample ff_sig;
    t_sample c_fb1;
    t_sample c_fb2;
    t_sample c_ff1;
    t_sample c_ff2;
    t_sample c_ff3;
} t_biquadctl;

typedef struct vaf_tilde
{
    t_object x_obj;
    t_float x_f;
    t_biquadctl x_cspace;
    t_biquadctl *x_ctl;
    t_sample prev_sample;
} t_vaf_tilde;

t_class *vaf_tilde_class;

static void vaf_tilde_list(t_vaf_tilde *x, t_symbol *s, int argc, t_atom *argv);

static void *vaf_tilde_new(t_symbol *s, int argc, t_atom *argv)
{
    t_vaf_tilde *x = (t_vaf_tilde *)pd_new(vaf_tilde_class);
    outlet_new(&x->x_obj, &s_signal);
    x->x_ctl = &x->x_cspace;
    x->x_cspace.c_x1 = x->x_cspace.c_x2 = 0;
    x->x_cspace.ff_sig = 0;
    vaf_tilde_list(x, s, argc, argv);
    x->x_f = 0;
    x->prev_sample = 0;
    return (x);
}

static t_int *vaf_tilde_perform(t_int *w)
{
    t_sample *in = (t_sample *)(w[1]);
    t_sample *out = (t_sample *)(w[2]);
    t_biquadctl *c = (t_biquadctl *)(w[3]);
    int n = (t_int)(w[4]);
    int i;
    t_sample ff_sig = c->ff_sig;
    t_sample last = c->c_x1;
    t_sample prev = c->c_x2;
    t_sample fb1 = c->c_fb1;
    t_sample fb2 = c->c_fb2;
    t_sample ff1 = c->c_ff1;
    t_sample ff2 = c->c_ff2;
    t_sample ff3 = c->c_ff3;
    for (i = 0; i < n; i++)
    {
        *out = 2*tanh(*in++ + ff_sig);
        t_sample output =  *out++ + fb1 * last + fb2 * prev; //Direct Form II implementation
        if (PD_BIGORSMALL(output))
            output = 0;
        ff_sig = ff1 * output + ff2 * last + ff3 * prev;
        prev = last;
        last = output;
    }
    c->ff_sig = ff_sig;
    c->c_x1 = last;
    c->c_x2 = prev;
    return (w+5);
}

static void vaf_tilde_list(t_vaf_tilde *x, t_symbol *s, int argc, t_atom *argv)
{
    t_float freq = atom_getfloatarg(0, argc, argv);
    t_float Bw = atom_getfloatarg(1, argc, argv);
    t_float G = atom_getfloatarg(2, argc, argv);
    t_float fs = sys_getsr(); //usare sys_getsr()
    
    t_float w = 2*PI*freq/fs;
    t_float g = (float) exp10f((float)G/20);
    t_float Q = 2*freq/Bw;
    t_float k = (4/(1+g))*tan(w/(2*Q));
    
    t_float ff1 = 1;
    t_float ff2 = -2*cos(w)/(1+k*g);
    t_float ff3 = (1-k*g)/(1+k*g);
    
    t_float fb1 = -2*cos(w)/(1+k);
    t_float fb2 = (1-k)/(1+k);
    t_float den1 = fb1;
    t_float den2 = fb2;
    fb1 = -ff2; // Il segno meno serve per far interpretare correttamente i coefficienti a PD
    fb2 = -ff3; // Il segno meno serve per far interpretare correttamente i coefficienti a PD
    ff1 = 0;
    ff2 = ff2 - den1;
    ff3 = ff3 - den2;
    
    post("BPF coeff %f %f %f %f %f",fb1,fb2,ff1,ff2,ff3);
    
    t_float discriminant = fb1 * fb1 + 4 * fb2;
    t_biquadctl *c = x->x_ctl;
    if (discriminant < 0) /* imaginary roots -- resonant filter */
    {
        /* they're conjugates so we just check that the product
         is less than one */
        if (fb2 >= -1.0f) goto stable;
    }
    else    /* real roots */
    {
        /* check that the parabola 1 - fb1 x - fb2 x^2 has a
         vertex between -1 and 1, and that it's nonnegative
         at both ends, which implies both roots are in [1-,1]. */
        if (fb1 <= 2.0f && fb1 >= -2.0f &&
            1.0f - fb1 -fb2 >= 0 && 1.0f + fb1 - fb2 >= 0)
            goto stable;
    }
    /* if unstable, just bash to zero */
    fb1 = fb2 = ff1 = ff2 = ff3 = 0;
    post("Filter UNSTABLE");
stable:
    c->c_fb1 = fb1;
    c->c_fb2 = fb2;
    c->c_ff1 = ff1;
    c->c_ff2 = ff2;
    c->c_ff3 = ff3;
}

static void vaf_tilde_set(t_vaf_tilde *x, t_symbol *s, int argc, t_atom *argv)
{
    t_biquadctl *c = x->x_ctl;
    c->c_x1 = atom_getfloatarg(0, argc, argv);
    c->c_x2 = atom_getfloatarg(1, argc, argv);
    c->ff_sig = atom_getfloatarg(2,argc,argv);
}

static void vaf_tilde_dsp(t_vaf_tilde *x, t_signal **sp)
{
    dsp_add(vaf_tilde_perform, 4,
            sp[0]->s_vec, sp[1]->s_vec,
            x->x_ctl, sp[0]->s_n);
    
}

void vaf_tilde_setup(void)
{
    vaf_tilde_class = class_new(gensym("vaf~"), (t_newmethod)vaf_tilde_new,
                                0, sizeof(t_vaf_tilde), 0, A_GIMME, 0);
    CLASS_MAINSIGNALIN(vaf_tilde_class, t_vaf_tilde, x_f);
    class_addmethod(vaf_tilde_class, (t_method)vaf_tilde_dsp, gensym("dsp"), 0);
    class_addlist(vaf_tilde_class, vaf_tilde_list);
    class_addmethod(vaf_tilde_class, (t_method)vaf_tilde_set, gensym("set"),
                    A_GIMME, 0);
    class_addmethod(vaf_tilde_class, (t_method)vaf_tilde_set, gensym("clear"),
                    A_GIMME, 0);
}
